using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using Microsoft.Xna.Framework.Content;

namespace XNACS1Lib
{
    /// <summary>
    /// This is the base class for programming assignments.
    /// </summary>
    internal class XNACS1LibDrawHelper
    {
        private TextureTable m_TexTable;
        private XNACS1LibFontManager m_FontManager;
        private Vector2 m_WorldMin, m_WorldMax;
        private Vector2 m_DeviceSize;

        private Matrix mW2D;

        private GraphicsDevice m_GraphicsDevice = null;
        private SpriteBatch mSpriteBatch = null;

        private Texture2D m_TheCircle = null;
        private Texture2D m_TheSquare = null;

        private List<XNACS1Primitive> m_DrawSet =null;

        private bool m_ZuneLandscape = false;

        public XNACS1LibDrawHelper(GraphicsDevice graphicsDevice, ContentManager c, XNACS1Rectangle worldBound, XNACS1LibFontManager fontManager, Vector2 deviceSize)
        {
            m_DrawSet = new List<XNACS1Primitive>();
            
            m_GraphicsDevice = graphicsDevice;

            mSpriteBatch = new SpriteBatch(m_GraphicsDevice);

            //
            SetTransformationMatrices(worldBound, deviceSize);
            
            m_TexTable = new TextureTable(c);
            m_FontManager = fontManager;
            fontManager.SetSpriteBatch(mSpriteBatch);

            m_TheCircle = FindTexture("XNALibCircle");
            m_TheSquare = FindTexture("XNALibSquare");
        }

        public void SetTransformationMatrices(XNACS1Rectangle worldBound, Vector2 deviceSize)
        {

            Vector2 wMin = worldBound.MinBound;
            Vector2 wMax = worldBound.MaxBound;

            Vector2 size = wMax - wMin;
            Matrix toOrg, toDC;
            toOrg = Matrix.CreateTranslation(-wMin.X, -wMin.Y, 0.0f);
            toDC = Matrix.CreateScale(deviceSize.X / size.X, deviceSize.Y / size.Y, 1.0f);
            mW2D = toOrg * toDC;
            
            m_WorldMin = wMin;
            m_WorldMax = wMax;
            m_DeviceSize = deviceSize;
        }

        public void SetZuneLandscape(bool toLandscape)
        {
            m_ZuneLandscape = toLandscape;
        }

        public Texture2D FindTexture(String name)
        {
            Texture2D tex = null;
            if (null != name)
            {
                tex = m_TexTable.FindTexture(name);
                if (tex == null)
                {
                    m_TexTable.LoadTexture(name);
                    tex = m_TexTable.FindTexture(name);
                }
            }
            return tex;
        }


        //private Matrix ComputeXform(Vector2 scale, float rotateInDegree, Vector2 translate, Vector2 pivot)
        //{
        //    Matrix pm, tm, sm, rm;
        //    pm = Matrix.CreateTranslation(-pivot.X, -pivot.Y, 0.0f);
        //    tm = Matrix.CreateTranslation(translate.X + pivot.X, translate.Y + pivot.Y, 0.0f);
        //    rm = Matrix.CreateRotationZ((float)(rotateInDegree * (Math.PI / 180.0f)));
        //    sm = Matrix.CreateScale(scale.X, scale.Y, 1.0f);
        //    return pm * sm * rm * tm;
        //}

        //private void XformVertices(Matrix m, VertexPositionColorTexture[] input, VertexPositionColorTexture[] output)
        //{
        //    for (int i = 0; i < input.Length; i++)
        //    {
        //        // output[i] = input[i];
        //        output[i].Position = Vector3.Transform(input[i].Position, m);
        //        output[i].Position = Vector3.Transform(output[i].Position, mW2D);
        //        output[i].Position.Y = m_DeviceSize.Y - output[i].Position.Y;   // drawing origin is top/left
        //    }
        //}

        private Rectangle World2Device(Vector3 tl, float w, float h)
        {
            Vector3 br = new Vector3(tl.X + w, tl.Y - h, 0f);
            tl = Vector3.Transform(tl, mW2D);
            br = Vector3.Transform(br, mW2D);
            w = br.X - tl.X;
            h = tl.Y - br.Y;

            Rectangle r;
#if ZUNE
            if (m_ZuneLandscape) {
                int Xdc = (int)(240f - tl.Y + 0.5f);
                int Ydc = (int)(320f - w - tl.X + 0.5f);
                r = new Rectangle(Xdc, Ydc, (int)(w + 0.5f), (int)(h + 0.5f));
            } else {
#endif
                tl.Y = m_DeviceSize.Y - tl.Y;
                r = new Rectangle((int)(tl.X+0.5f), (int)(tl.Y+0.5f), (int)(w+0.5f), (int)(h+0.5f));

#if ZUNE
            }
#endif

            return r;
        }

        /// <summary>
        /// Compute the top-left position given that XNA sprite rotation is always top-left.
        /// Our rotation is middle of geometry
        /// 
        // desiredRotationCenterX + px = topLeft.X 
        // desiredRotationCenterY + py = topLeft.Y 
        /// 
        /// <returns></returns>
        /// </summary>
        private Vector3 ComputeTopLeft(Vector3 tl, float px, float py, float r)
        {
            double xS = px * Math.Sin(r);
            double xC = px * Math.Cos(r);
            double yS = py * Math.Sin(r);
            double yC = py * Math.Cos(r);
            return new Vector3(tl.X - px + (float)(xC - yS), 
                               tl.Y - py + (float)(xS + yC),
                               0f);
        }

        /// <summary>
        /// Actually draws the primitive onto screen.
        /// </summary>
        /// <param name="dcRec">The dcRec to draw</param>
        /// <param name="texName">Texture name</param>
        /// <param name="useTex">if texName is null, use this texture instead (TheCircle or TheSquare)</param>
        /// <param name="color">color to draw in in the absence of texName</param>
        /// <param name="rotation">to rotate</param>
        private void DrawDCRectangle(Rectangle dcRec, String texName, Texture2D useTex, Color color, float rotation)
        {
            Texture2D tex = FindTexture(texName);
            Color useColor = color;

            if (null == tex)
            {
                tex = useTex;
            }
            else
            {
                useColor = Color.White;
            }

#if ZUNE
            if (m_ZuneLandscape) 
                rotation -= 180;
#endif

            mSpriteBatch.Draw(tex, dcRec, null, useColor, -rotation, Vector2.Zero, SpriteEffects.None, 1f);
        }

        /// <summary>
        /// Draws a circle.
        /// </summary>
        /// <param name="center">Center of the circle.</param>
        /// <param name="radius">Radius of the circle.</param>
        /// <param name="rotation">Rotation of the circle (about the center, in degree).</param>
        /// <param name="color">Color of the circle.</param>
        /// <param name="texName">Name of texture (use null for none) for the circle.</param>
        public void DrawCircle(Vector2 center, float radius, float rotation, Color color, String texName)
        {
            Vector3 topLeft = new Vector3(center.X - radius, center.Y + radius, 0f);
            
            // convert rotation to radian
            rotation = MathHelper.ToRadians(rotation);

            // desiredRotationCenterX + px = topLeft.X 
            // desiredRotationCenterY + py = topLeft.Y
            topLeft = ComputeTopLeft(topLeft, -radius, radius, rotation);
            
            float width = 2f * radius;
            Rectangle dcRec = World2Device(topLeft, width, width);

            DrawDCRectangle(dcRec, texName, m_TheCircle, color, rotation);
        }

        /// <summary>
        /// Draws a roated rectangle.
        /// </summary>
        /// <param name="center">Center of the rectangle.</param>
        /// <param name="xWidth">Horizontal width of the rectangle.</param>
        /// <param name="yHeight">Vertical height of the rectangle.</param>
        /// <param name="rot">Rotation in degrees.</param>
        /// <param name="color">Color of the rectangle.</param>
        /// <param name="texName">Name of texture (use null for none) for the rectangle.</param>
        public void DrawRectangle(Vector2 center, float xWidth, float yHeight, float rot, Color color, String texName)
        {
            Vector3 topLeft = new Vector3(center.X - (xWidth / 2f), center.Y + (yHeight / 2f), 0f);

            // convert rotation to rad
            rot = MathHelper.ToRadians(rot);

            // desiredRotationCenterX + px = topLeft.X 
            // desiredRotationCenterY + py = topLeft.Y
            topLeft = ComputeTopLeft(topLeft, -xWidth/2f, yHeight/2f, rot);

            Rectangle dcRec = World2Device(topLeft, xWidth, yHeight);
            DrawDCRectangle(dcRec, texName, m_TheSquare, color, rot);
        }

        public void DrawLineSegments(Vector2 a, Vector2 b)
        {
            //VertexPositionColorTexture[] v = new VertexPositionColorTexture[2];
            //VertexPositionColorTexture[] vo = new VertexPositionColorTexture[2];
            //v[0] = new VertexPositionColorTexture(
            //            new Vector3(a.X, a.Y, 0.0f), Color.White, new Vector2(0.0f, 0.0f));
            //v[1] = new VertexPositionColorTexture(
            //            new Vector3(b.X, b.Y, 0.0f), Color.White, new Vector2(1.0f, 1.0f));
            //Matrix m = ComputeXform(new Vector2(1.0f, 1.0f), 0.0f, new Vector2(0.0f, 0.0f), new Vector2(0.0f, 0.0f));
            //XformVertices(m, v, vo);
            //m_GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, vo, 0, 1);
        }

        public void BeginDraw()
        {
            mSpriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.None);
        }

        public void EndDraw()
        {
            mSpriteBatch.End();
        }

        private Vector2 WCToDC(Vector2 wc)
        {
            Vector2 dc = new Vector2();
            Vector2 wcSize = m_WorldMax - m_WorldMin;
            dc.X = (wc.X - m_WorldMin.X) * m_DeviceSize.X / wcSize.X;
            dc.Y = ((wc.Y - m_WorldMin.Y) * m_DeviceSize.Y / wcSize.Y);
            dc.Y = m_DeviceSize.Y - dc.Y;
            return dc;
        }

        public void DrawFontAt(Vector2 wcAt, String msg, Color c)
        {
            m_FontManager.DrawFontsAt(WCToDC(wcAt), msg, c);
        }

        #region draw set support
        public void AddToSet(XNACS1Primitive p)
        {
            if (null != m_DrawSet)
            {
                if (!m_DrawSet.Contains(p))
                    m_DrawSet.Add(p);
            }
        }

        public void RemoveFromSet(XNACS1Primitive p)
        {
            if (null != m_DrawSet)
                m_DrawSet.Remove(p);
        }

        public void RemoveAllFromSet()
        {
            if (null != m_DrawSet)
                m_DrawSet.Clear();
        }

        public void TopOfDrawSet(XNACS1Primitive p)
        {
            if (null != m_DrawSet)
            {
                if (IsInDrawSet(p))
                {
                    m_DrawSet.Remove(p);
                    m_DrawSet.Add(p);
                }
            }
        }

        public void DrawAll()
        {
            if (null != m_DrawSet)
            {
                foreach (XNACS1Primitive p in m_DrawSet)
                    p.Draw();
            }
        }

        public void TravelAll()
        {
            foreach (XNACS1Primitive p in m_DrawSet)
                p.TravelPrimitive();

        }

        public bool IsInDrawSet(XNACS1Primitive p)
        {
            return m_DrawSet.Contains(p);
        }

        public int DrawSetSize()
        {
            return m_DrawSet.Count;
        }
        #endregion draw set support
        
    }
}